



if not exist %SP_DIR%/poetry_plugin_export/__init__.py exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip show poetry-plugin-export | rg -F "Version: 1.3.0"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pip check | rg -F "requires poetry, which is not installed."
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
